unit UMainSys;

interface

uses
	Windows, USmDx3D, DDUtils, DDVer, UTileMap, USmPerson, USmUtil;

const
	MAP_X_WIDE    = 4;
	MAP_Y_WIDE    = 5;
	TILE_X_SIZE   = 60;
	TILE_X_HALF   = TILE_X_SIZE div 2;
	TILE_Y_SIZE   = 41;
	TILE_Y_HALF   = 14;
	TILE_Y_HEIGHT = 13;

	TILE_X_AREA   = TILE_X_SIZE;
	TILE_Y_AREA   = 45;

	MAX_IMAGE     = 2;
	MAX_TEXTURE   = 10;
	MAX_OBJECT    = 100;

	MAX_BITMAP_TILE     = 4 * 7;
	MAX_CHARA_TILE      = MAX_CHARA_FRAME * MAX_CHARA_DIRECTION;
	MAX_CHARA_NUMBER    = 4;

	MAIN_CHARACTER : Integer = 0;

type
	TGameOption = record
		Handle        : THandle;
		Width         : Integer;
		Height        : Integer;
		Depth         : Integer;
		UseFullScreen : Boolean;
	end;

	TBitmapTile = record
		Image         : Longint;
		Height        : Word;
		ARect         : TRect;
		HotSpot       : TPoint;
	end;

	TMainSystem = packed class
		constructor Create( Handle : THandle );
		procedure   Free;

	private
		function  Map2AbsCoordX( xMap, yMap : Integer ) : Integer;
		function  Map2AbsCoordY( xMap, yMap : Integer ) : Integer;
		function  Abs2MapCoordX( xAbs, yAbs : Integer ) : Integer;
		function  Abs2MapCoordY( xAbs, yAbs : Integer ) : Integer;

	public
		BitmapTile : array[0..MAX_BITMAP_TILE] of TBitmapTile;
		CharaTile  : array[0..Pred( MAX_CHARA_NUMBER ), 0..Pred( MAX_CHARA_TILE ) ] of TBitmapTile;
		BMPImage   : array[0..Pred( MAX_IMAGE   )] of II_DirectDrawSurface;
		Objects    : array[0..Pred( MAX_OBJECT  )] of TMirage;
		TileMap    : TTileMap;
		KeyList    : array[0..0] of TKeyList;

		procedure Abs2MapCoord ( xAbs, yAbs : Integer; var xMap, yMap : Integer );
		function  GetFreeObjectNumber : Integer;
		function  CreateSprite( x, y : Integer; vx, vy, ax, ay : Single; TileType : ETileType; Image : Integer; Angle : Integer; LifeTime : Integer ) : Boolean;
		procedure DrawQuaterViewEx( xViewPos, yViewPos : Integer );
		function  MainProc : Boolean;

	end;

var
	MainSystem  : TMainSystem;
	SmDirect3D  : TSmDirect3D;
	GameOption  : TGameOption;
	GameCursor  : TPoint;
	GameKey     : Word;
	ResourceDLL : integer = 0;

implementation

var
	CURSOR_MAP_POS : TPoint = ( x : 0; y : 0 );
	CurrentTile    : Integer;
	CurrentHeight  : Integer;


procedure   MakeBitmapTile( out BitmapTile : TBitmapTile; Image : Integer; ARect : TRect; HotSpot : TPoint );
begin
	BitmapTile.Image   := Image;
	BitmapTile.ARect   := ARect;
	BitmapTile.Height  := 0;
	BitmapTile.HotSpot := HotSpot;
end;

{ TMainSystem }

constructor TMainSystem.Create( Handle : THandle );
const
	HOT_SPOT : array[0..Pred( MAX_CHARA_NUMBER ), 0..Pred( MAX_CHARA_DIRECTION ),1..2] of Integer = (
		( ( 34, 55 ), ( 20, 55 ), ( 32, 54 ), ( 22, 54 ) ),
		( ( 26, 57 ), ( 28, 57 ), ( 27, 57 ), ( 27, 57 ) ),
		( ( 27, 57 ), ( 27, 57 ), ( 26, 58 ), ( 28, 58 ) ),
		( ( 34, 55 ), ( 20, 55 ), ( 32, 54 ), ( 22, 54 ) )
	);
var
	i, j, k : Integer;
begin
	Inherited Create;

	// Ķ ʱȭ
	ZeroMemory( @BMPImage, sizeof( BMPImage ) );
	ZeroMemory( @Objects,  sizeof( Objects  ) );

	GameOption.Handle := Handle;

	// Direct X ʱȭ
	with GameOption do
		SmDirect3D    := TSmDirect3D.Create3D( Handle, Width, Height, Depth, UseFullScreen );

	// Image ε
	BMPImage[0] := DDCreateSurface( SmDirect3D.DirectDraw, GameOption.Width, GameOption.Height, FALSE );
	DDReLoadPCX( BMPImage[0], 'Neto1' );
	BMPImage[1] := DDCreateSurface( SmDirect3D.DirectDraw, GameOption.Width, GameOption.Height, FALSE );
	DDReLoadPCX( BMPImage[1], 'Neto2' );

	//  б
	TileMap := TTileMap.Create;
	TileMap.RandomMap( 100, 100 );

	// BitmapTile   Ҵ
	for j := 0 to Pred(MAX_BITMAP_TILE div 4) do begin
		for i := 0 to 3 do begin
			BitmapTile[j*4+i].Image   := 0;
			BitmapTile[j*4+i].ARect   := Bounds( i * TILE_X_AREA, TILE_Y_AREA*j + 1, TILE_X_AREA, TILE_Y_AREA );
			BitmapTile[j*4+i].Height  := 0;
			BitmapTile[j*4+i].HotSpot := Point( 0, 0 );
		end;
	end;

	// CharaTile   Ҵ
	for k := 0 to Pred( MAX_CHARA_NUMBER ) do begin
		for j := 0 to Pred( MAX_CHARA_DIRECTION ) do begin
			for i := 0 to Pred( MAX_CHARA_FRAME ) do begin
				MakeBitmapTile(	CharaTile[k][j*MAX_CHARA_FRAME+i], 1, Bounds( 55*(MAX_CHARA_FRAME*k+i)+1, 71*j+1, 54, 70 ), Point( HOT_SPOT[k][j][1], HOT_SPOT[k][j][2] ) );
			end;
		end;
	end;

	// Ű 
	KeyList[0][SM_KEY_LEFT  ] := VK_LEFT;
	KeyList[0][SM_KEY_RIGHT ] := VK_RIGHT;
	KeyList[0][SM_KEY_UP    ] := VK_UP;
	KeyList[0][SM_KEY_DOWN  ] := VK_DOWN;
	KeyList[0][SM_KEY_SHOOT ] := VK_CONTROL;
	KeyList[0][SM_KEY_CHANGE] := VK_SPACE;
	KeyList[0][SM_KEY_LOCK  ] := VK_SHIFT;

	// ĳ 
	Objects[MAIN_CHARACTER] := TPersonManualLight.Create( MAIN_CHARACTER, 0, Map2AbsCoordX( 1, 1 ), Map2AbsCoordY( 1, 1 ) );
//	Objects[1] := TPerson.Create( 1, 1, Map2AbsCoordX( 2, 2 ), Map2AbsCoordY( 2, 2 ) );
end;

procedure   TMainSystem.Free;
var
	i : Integer;
begin
	TileMap.Free;

	for i := 0 to Pred( MAX_IMAGE ) do begin
		if Assigned( BMPImage[i] ) then
			BMPImage[i] := nil;
	end;

	SmDirect3D.Free;

	Inherited;
end;

function    TMainSystem.GetFreeObjectNumber : Integer;
var
	i : Integer;
begin
	for i := 0 to Pred( MAX_OBJECT ) do begin
		if not Assigned( Self.Objects[i] ) then begin
			GetFreeObjectNumber := i;
			Exit;
		end;
	end;

	GetFreeObjectNumber := -1;
end;

function    TMainSystem.CreateSprite( x, y : Integer; vx, vy, ax, ay : Single; TileType : ETileType; Image : Integer; Angle : Integer; LifeTime : Integer ) : Boolean;
var
	Index : Integer;
begin
	CreateSprite := FALSE;

	Index := GetFreeObjectNumber;

	if Index < 0 then
		Exit;

	Self.Objects[Index] := TSprite.Create( Index, x, y, vx, vy, ax, ay, TileType, Image, Angle, LifeTime );

	CreateSprite := TRUE;
end;

function  TMainSystem.Map2AbsCoordX( xMap, yMap : Integer ) : Integer;
begin
	Map2AbsCoordX := ( ( xMap + yMap ) * TILE_X_HALF );
end;

function  TMainSystem.Map2AbsCoordY( xMap, yMap : Integer ) : Integer;
begin
	Map2AbsCoordY := ( ( ( TileMap.MapHeader.MapSize.x - xMap ) + yMap ) * TILE_Y_HALF );
end;

function  TMainSystem.Abs2MapCoordX( xAbs, yAbs : Integer ) : Integer;
begin
	Abs2MapCoordX := ( ( xAbs div TILE_X_HALF + ( TileMap.MapHeader.MapSize.x - yAbs div TILE_Y_HALF ) ) div 2 );
end;

function  TMainSystem.Abs2MapCoordY( xAbs, yAbs : Integer ) : Integer;
begin
	Abs2MapCoordY := ( ( xAbs div TILE_X_HALF - ( TileMap.MapHeader.MapSize.x - yAbs div TILE_Y_HALF ) ) div 2 );
end;

procedure TMainSystem.Abs2MapCoord( xAbs, yAbs : Integer; var xMap, yMap : Integer );
const
	GRAD : Double = TILE_Y_HALF * 1.0 / TILE_X_HALF;
var
	px, py, lx, ly, tx, ty : Integer;
begin

	(*  ǥ ǥ (0,0)    ǥ  *)
	px := xAbs + TILE_X_SIZE + 1;
	py := yAbs + TILE_Y_SIZE * 2 - TILE_Y_HEIGHT;

	tx := px div ( TILE_X_HALF * 2 );
	ty := py div ( TILE_Y_HALF * 2 );

	lx := ( tx - ty );
	ly := ( tx + ty );

	tx := px mod ( TILE_X_HALF * 2 );
	ty := py mod ( TILE_Y_HALF * 2 );

	if ( GRAD * tx + ty <  TILE_Y_HALF   ) then Dec( ly );
	if ( GRAD * tx + ty >  TILE_Y_HALF*3 ) then Inc( ly );
	if ( GRAD * tx - ty >  TILE_Y_HALF   ) then Inc( lx );
	if ( GRAD * tx - ty < -TILE_Y_HALF   ) then Dec( lx );

	xMap := lx + TileMap.MapHeader.MapSize.x div 2;
	yMap := ly - TileMap.MapHeader.MapSize.x div 2;
	xMap := lx + 100 div 2;
	yMap := ly - 100 div 2;
end;

procedure   TMainSystem.DrawQuaterViewEx( xViewPos, yViewPos : Integer );
var
	i, x, y, z                    : Integer;
	xMapOrg,  yMapOrg             : Integer;
	xMapPos,  yMapPos             : Integer;
	xMapCur,  yMapCur             : Integer;
	xPosOrg,  yPosOrg             : Integer;
	xTilePos, yTilePos, zTilePos  : Integer;
	xPitch,   yPitch              : Integer;
	_Tile, _Height, _Person, _Aux : Integer;
	MA                            : TMapAttribute;
	TempRect                      : TRect;
	AvailablePersonList           : array[0..Pred( MAX_OBJECT )] of Boolean;
var
	DC    : HDC;
	a, b  : String;
const
	PreTick : Integer = -1;
	CurTick : Integer = 0;
	Tick    : Integer = 0;

begin
	ZeroMemory( @AvailablePersonList, sizeof( AvailablePersonList ) );
	for i := 0 to Pred( MAX_OBJECT ) do begin
		if Assigned( Objects[i] ) then
			AvailablePersonList[i] := TRUE;
	end;

	Dec( xViewPos, ( GameOption.Width  div 2 ) + TILE_X_SIZE div 2 );
	Dec( yViewPos, ( GameOption.Height div 2 ) + TILE_Y_SIZE );

	xMapOrg := Abs2MapCoordX( xViewPos, yViewPos );
	yMapOrg := Abs2MapCoordY( xViewPos, yViewPos );

	xPosOrg := -TILE_X_HALF;
	yPosOrg := -TILE_Y_HALF;

	xPitch  := Map2AbsCoordX( xMapOrg, yMapOrg ) - xViewPos;
	yPitch  := Map2AbsCoordY( xMapOrg, yMapOrg ) - yViewPos;

	Abs2MapCoord( xViewPos + GameCursor.x, yViewPos + GameCursor.y, xMapCur, yMapCur );

	CURSOR_MAP_POS.x := xMapCur;
	CURSOR_MAP_POS.y := yMapCur;

	for y := 0 to ( GameOption.Height div TILE_Y_HALF ) + 20 do begin
		xTilePos  := xPosOrg + 0;
		yTilePos  := yPosOrg + y * TILE_Y_HALF - TILE_Y_HALF;

		Inc( xTilePos, xPitch );
		Inc( yTilePos, yPitch );

		xMapPos   := xMapOrg - y div 2;
		yMapPos   := yMapOrg + (y+1) div 2;

		// 
		if ( y mod 2 ) > 0 then
			Inc( xTilePos, TILE_X_HALF );

		for x := 0 to ( GameOption.Width div TILE_X_HALF ) + 1 do begin

			_Tile   := TileMap[maTile  , xMapPos, yMapPos];
			_Height := TileMap[maHeight, xMapPos, yMapPos];
			_Person := TileMap[maPerson, xMapPos, yMapPos];

			for z := 0 to 0 do
				SmDirect3D.PutSprite( xTilePos, yTilePos - z*TILE_Y_HEIGHT, BMPImage[BitmapTile[_Tile].Image], BitmapTile[_Tile].ARect, smCentered );

			// Ŀ ġ  Ŀ
			if ( xMapCur = xMapPos ) and ( yMapCur = yMapPos ) then begin
				TempRect.Left   := 351;
				TempRect.Top    := 16;
				TempRect.Right  := 415;
				TempRect.Bottom := 39;
				SmDirect3D.PutSprite( xTilePos, yTilePos-16, BMPImage[0], TempRect, smCentered );
			end;

			Inc( xTilePos, TILE_X_SIZE );

			Inc( xMapPos );
			Inc( yMapPos );
		end;
	end;

	for i := 0 to Pred( MAX_OBJECT ) do begin
		if Assigned( Objects[i] ) and ( Objects[i] is TPerson ) then begin
			( Objects[i] as TPerson ).DisplaySelf( -xViewPos, -yViewPos, TRUE );
		end;
	end;

	// Ŀ ǥ
	TempRect.Left   := 251;
	TempRect.Top    := 16;
	TempRect.Right  := 269;
	TempRect.Bottom := 34;
	SmDirect3D.PutSprite( GameCursor.x, GameCursor.y, BMPImage[0], TempRect, smLeftTop );

	// ǥ 
	SmDirect3D.DirectDrawSurface.GetDC( DC );

	if ( PreTick >= 0 ) then begin
		CurTick := timeGetTime;
		Tick    := CurTick - PreTick;
		PreTick := CurTick;
		if Tick <> 0 then begin
			str( 1000 / Tick : 5 : 2, a );
		end else begin
			a := '...';
		end;
		TextOut( DC, 0, 0, PChar(a), Length(a) );
	end else begin
		PreTick := timeGetTime;
	end;

	MainSystem.Abs2MapCoord( Objects[MAIN_CHARACTER].mX, Objects[MAIN_CHARACTER].mY, x, y );
	str( x : 3, a );
	str( y : 3, b );
	a := a + ' : ' + b;
	TextOut( DC, 0, 20, PChar(a), Length(a) );

	SmDirect3D.DirectDrawSurface.ReleaseDC( DC );
end;

function    TMainSystem.MainProc : Boolean;
const
	G_POS_X : Integer = -1;
	G_POS_Y : Integer = -1;
	MOVE_STEP_X = TILE_X_SIZE div 20;
	MOVE_STEP_Y = TILE_Y_SIZE div 20;

	__TEMP_ANGLE : Integer = 0;
	__TEMP_SCALE : Integer = 256;
	__TEMP_COUNT : Integer = 0;
	CharIndex    : Integer = 0;
var
	i : Integer;
begin
	if ( G_POS_X = -1 ) then begin
		G_POS_X := Map2AbsCoordX( 0, 0 );
		G_POS_Y := Map2AbsCoordY( 0, 0 );
	end;

	if GameCursor.x <= 0 then
		Inc( G_POS_X, -MOVE_STEP_X );

	if GameCursor.x >= Pred( GameOption.Width ) then
		Dec( G_POS_X, -MOVE_STEP_X );

	if GameCursor.y <= 0 then
		Inc( G_POS_Y, -MOVE_STEP_Y );

	if GameCursor.y >= Pred( GameOption.Height ) then
		Dec( G_POS_Y, -MOVE_STEP_Y );

	if HiByte( GetAsyncKeyState( VK_NUMPAD4  ) ) > 0 then
		Inc( G_POS_X, -MOVE_STEP_X );

	if HiByte( GetAsyncKeyState( VK_NUMPAD6 ) ) > 0 then
		Dec( G_POS_X, -MOVE_STEP_X );

	if HiByte( GetAsyncKeyState( VK_NUMPAD8 ) ) > 0 then
		Inc( G_POS_Y, -MOVE_STEP_Y );

	if HiByte( GetAsyncKeyState( VK_NUMPAD2 ) ) > 0 then
		Dec( G_POS_Y, -MOVE_STEP_Y );

	for i := 0 to Pred( MAX_OBJECT ) do begin
		if Assigned( Objects[i] ) then begin
			if not Objects[i].DoAction( i = MAIN_CHARACTER ) then begin
				Objects[i].Free;
				Objects[i] := nil;
			end;
		end;
	end;

	G_POS_X := Objects[MAIN_CHARACTER].mX;
	G_POS_Y := Objects[MAIN_CHARACTER].mY;

	DrawQuaterViewEx( G_POS_X, G_POS_Y );

	SmDirect3D.FlipPage;

	Result := ( HiByte( GetAsyncKeyState( VK_ESCAPE ) ) = 0 );
end;

end.

