unit UResManager;

interface

uses
	Windows;

type

	TSmStream = class
	private
	public
		IsAvailable : Boolean;
		
		function Read( var Buffer; Count : Longint ) : Longint; virtual; abstract;
		function Seek( Offset : Longint; Origin : Word) : Longint; virtual; abstract;
	end;

	TSmFileStream = class( TSmStream )
	private
		FFile            : file;
	public
		constructor Create( const FileName : string );
		destructor  Destroy; override;
		function    Read( var Buffer; Count : Longint ) : Longint; override;
		function    Seek( Offset : Longint; Origin : Word) : Longint; override;
	end;


	TSmResourceStream = class( TSmStream )
	private
		FMemory    : Pointer;
		FSize      : Longint;
		FPosition  : Longint;
		HResInfo   : HRSRC;
		HGlobal    : THandle;
		procedure   Initialize( Instance : THandle; Name, ResType : PChar );
	public
		constructor Create( Instance : THandle; const ResName : string; ResType : PChar );
		destructor  Destroy; override;
		function    Read( var Buffer; Count : Longint ) : Longint; override;
		function    Seek( Offset : Longint; Origin : Word) : Longint; override;
  end;

implementation

constructor TSmResourceStream.Create( Instance : THandle; const ResName : string; ResType : PChar );
begin
	inherited Create;

	IsAvailable := FALSE;

	Initialize( Instance, PChar( ResName ), ResType );
end;

procedure TSmResourceStream.Initialize( Instance : THandle; Name, ResType : PChar );
begin
	HResInfo   := FindResource( Instance, Name, ResType );
	if HResInfo = 0 then
		Exit;

	HGlobal    := LoadResource( Instance, HResInfo );
	if HGlobal = 0 then
		Exit;

	FMemory    := LockResource( HGlobal );
	FSize      := SizeOfResource( Instance, HResInfo );

	IsAvailable := TRUE;
end;

destructor TSmResourceStream.Destroy;
begin
	if ( HGlobal <> 0 ) then begin
		UnlockResource( HGlobal );
		FreeResource( HGlobal );
	end;

	inherited Destroy;
end;

function TSmResourceStream.Read( var Buffer; Count : Longint) : Longint;
begin
	if not IsAvailable then begin
		Result := 0;
		Exit;
	end;

	if ( FPosition >= 0 ) and ( Count >= 0 ) then begin
		Result := FSize - FPosition;
		if Result > 0 then begin
			if Result > Count then
				Result := Count;
			Move( Pointer( Longint( FMemory ) + FPosition )^, Buffer, Result );
			Inc( FPosition, Result );
			Exit;
		end;
	end;
	Result := 0;
end;

function TSmResourceStream.Seek( Offset : Longint; Origin : Word) : Longint;
begin
	if not IsAvailable then begin
		Result := 0;
		Exit;
	end;

	case Origin of
		0 : FPosition := Offset;
		1 : Inc(FPosition, Offset);
		2 : FPosition := FSize + Offset;
	end;
	Result := FPosition;
end;

{ TSmFileStream }

constructor TSmFileStream.Create( const FileName : string );
begin
	Inherited Create;

	AssignFile( FFile, FileName );
 {$I-}
	Reset( FFile, 1 );
 {$I+}
	IsAvailable := ( IOResult = 0 );
end;

destructor TSmFileStream.Destroy;
begin
	if IsAvailable then
		CloseFile( FFile );

	inherited;
end;

function TSmFileStream.Read( var Buffer; Count : Integer ) : Longint;
var
	ActualRead : Longint;
begin
	if not IsAvailable then begin
		Result := 0;
		Exit;
	end;

	BlockRead( FFile, Buffer, Count, ActualRead );

	Result := ActualRead;
end;

function TSmFileStream.Seek( Offset : Integer; Origin : Word) : Longint;
begin
	if not IsAvailable then begin
		Result := 0;
		Exit;
	end;

	case Origin of
		1 : Offset := System.FilePos( FFile ) + Offset;
		2 : Offset := System.FileSize( FFile ) + Offset;
	end;

	System.Seek( FFile, Offset );

	Result := Offset;
end;

end.
