unit USmPerson;

interface

uses
	Windows, USmDx3D, USmUtil;

const
	MAX_CHARA_DIRECTION = 4;
	MAX_CHARA_FRAME     = 3;

type

	ETileType = ( ttETile, ttTexture );

	TMirage = class

		constructor Create( Number : Integer; x, y : Integer; TileType : ETileType; Image : Integer; LifeTime : Integer  );

	protected
		m_number    : Integer;
		m_x         : Integer;
		m_y         : Integer;
		m_sx        : Single;
		m_sy        : Single;
		m_tile_type : ETileType;
		m_image     : Integer;
		m_life_time : Integer;

	public
		procedure   WarpXY( x, y : Integer );                                    virtual;
		procedure   DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean ); virtual;
		function    DoAction( IsMain :  Boolean ) : Boolean;                     virtual;
		procedure   MoveXYAuto;                          virtual; abstract;
		function    MoveXY( x1, y1 : Single ) : Boolean; virtual; abstract;

		property    mX : Integer  read m_x  write m_x;
		property    mY : Integer  read m_y  write m_y;

	end;

	TSprite = class( TMirage )

		constructor Create( number : Integer; x, y : Integer; vx, vy, ax, ay : Single; TileType : ETileType; Image : Integer; Angle : Integer; LifeTime : Integer );
		procedure   Free;

	protected
		m_vx      : Single;
		m_vy      : Single;
		m_ax      : Single;
		m_ay      : Single;
		m_angle   : Integer;

	public
		procedure   MoveXYAuto;                                                  override;
		function    MoveXY( x1, y1 : Single ) : Boolean;                         override;
		procedure   DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean ); override;
		function    DoAction( IsMain :  Boolean ) : Boolean;                     override;

		property    mVX : Single  read m_vx  write m_vx;
		property    mVY : Single  read m_vy  write m_vy;
		property    mAX : Single  read m_ax  write m_ax;
		property    mAY : Single  read m_ay  write m_ay;

	end;

	TPerson = class( TMirage )

		constructor Create( number : Integer; ETileImage : Integer; x, y : Integer );

	protected
		m_direction    : Integer;
		m_circle_scale : Integer;
		m_face_count   : Integer;
		m_face_inc     : Integer;

		function    GetFaceImage : Integer;

	public

		procedure   MoveXYAuto;                                                  override;
		function    MoveXY     ( x1, y1 : Single ) : Boolean;                    override;
		procedure   DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean ); override;
		function    DoAction( IsMain :  Boolean ) : Boolean;                     override;

		procedure   DisplaySelfAfter( xDelta, yDelta : Integer; Selected : Boolean ); virtual;

		procedure   MoveXYEx   ( Angle : Integer; Distance : Single; Intelligence : Boolean );
		procedure   SetFace    ( Direction : Integer );
		function    GetFace   : Integer;
		procedure   IncFaceStep;

		procedure   SavePersonToFile  ( var f : file );
		procedure   LoadPersonFromFile( var f : file );

		property    mDirection    : Integer      read m_direction     write m_direction;
		property    mFaceCount    : Integer      read m_face_count    write m_face_count;
		property    mFaceInc      : Integer      read m_face_inc      write m_face_inc;

	end;

	TPersonManual = class( TPerson )

	public
		procedure   DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean );      override;
		function    DoAction( IsMain :  Boolean ) : Boolean;                          override;
		procedure   DisplaySelfAfter( xDelta, yDelta : Integer; Selected : Boolean ); override;

	end;

	TPersonManualLight = class( TPersonManual )

	public
		function    DoAction( IsMain :  Boolean ) : Boolean;                          override;

	end;

implementation

uses
	UMainSys, UTileMap;

{ TMirage }

constructor TMirage.Create( Number : Integer; x, y : Integer; TileType : ETileType; Image : Integer; LifeTime : Integer );
begin
	Self.WarpXY( x, y );

	m_number    := Number;
	m_tile_type := TileType;
	m_image     := Image;
	m_life_time := LifeTime;
end;

procedure   TMirage.WarpXY( x, y : Integer );
begin
	m_x  := x;
	m_y  := y;
	m_sx := x;
	m_sy := y;
end;

procedure   TMirage.DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean );
begin
	case m_tile_type of
		ttETile   :
		begin
		end;
		ttTexture :
		begin
		end;
	end;
end;

function    TMirage.DoAction( IsMain :  Boolean ) : Boolean;
begin
	DoAction := ( m_life_time > 0 );
	Dec( m_life_time );
end;

{ TSprite }

constructor TSprite.Create( number : Integer; x, y : Integer; vx, vy, ax, ay : Single; TileType : ETileType; Image : Integer; Angle : Integer; LifeTime : Integer );
begin
	Inherited Create( number, x, y, TileType, Image, LifeTime );

	m_vx    := vx;
	m_vy    := vy;
	m_ax    := ax;
	m_ay    := ay;
	m_angle := Angle;
end;

procedure   TSprite.Free;
begin
end;

procedure   TSprite.MoveXYAuto;
begin
	m_vx := m_vx + m_ax;
	m_vy := m_vy + m_ay;

	m_sx := m_sx + m_vx;
	m_sy := m_sy + m_vy;

	m_x  := Round( m_sx );
	m_y  := Round( m_sy );
end;

function    TSprite.MoveXY( x1, y1 : Single ) : Boolean;
begin
	m_sx := m_sx + x1;
	m_sy := m_sy + y1;

	m_x  := Round( m_sx );
	m_y  := Round( m_sy );

	MoveXY := TRUE;
end;

procedure   TSprite.DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean );
begin
	case m_tile_type of
		ttETile   :
		begin
		end;
		ttTexture :
		begin
		end;
   end;
end;

function    TSprite.DoAction( IsMain :  Boolean ) : Boolean;
begin
	MoveXYAuto;
	DoAction := Inherited DoAction( IsMain );
end;

{ TPerson }

constructor TPerson.Create( number : Integer; ETileImage : Integer; x, y : Integer );
begin
	Inherited Create( number, x, y, ttETile, ETileImage, 0 );

	m_direction    := 90;
	m_circle_scale := 256;
	m_face_inc     := 0;
	m_face_count   := 0;
end;

procedure TPerson.DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean );
begin
	IncFaceStep;

	with MainSystem, MainSystem.CharaTile[m_image][GetFaceImage+m_face_count] do
		SmDirect3D.PutSprite( m_x + xDelta - HotSpot.x, m_y + yDelta - HotSpot.y, BMPImage[Image], ARect, smLeftTop );
end;

procedure TPerson.DisplaySelfAfter( xDelta, yDelta : Integer; Selected : Boolean );
begin

end;

function  TPerson.DoAction( IsMain :  Boolean ) : Boolean;
begin
	if Random( 10 ) = 0 then
		SetFace( m_direction + 5 );

	MoveXY( SmCos( m_direction ) * 0.5, - SmSin( m_direction ) * 0.5 );

	DoAction := TRUE;
end;

procedure TPerson.IncFaceStep;
begin
	Inc( m_face_inc );
	m_face_count := (( m_face_inc div 30 ) mod 4);
	if m_face_count > 2 then
		m_face_count := 4 - m_face_count;
end;

procedure TPerson.LoadPersonFromFile( var f : file );
begin

end;

function  TPerson.MoveXY( x1, y1 : Single ) : Boolean;
var
//	xNew, yNew     : Integer;
	xMapOR, yMapOR : Integer;
	xMapDL, yMapDL : Integer;
	xMapDR, yMapDR : Integer;
	xMapUL, yMapUL : Integer;
	xMapUR, yMapUR : Integer;
	CheckRect      : TRect;
	Tile           : TRect;
	xLen, yLen     : Integer;
begin
	MoveXY := FALSE;

	Tile    := MainSystem.CharaTile[m_image][GetFaceImage+m_face_count].ARect;
	xLen    := Tile.Right - Tile.Left;
	yLen    := Tile.Bottom - Tile.Top;

	CheckRect.Left   := - xLen div 2;
	CheckRect.Right  := + xLen div 2;
	CheckRect.Top    := - 10;
	CheckRect.Bottom := + 10;

	MainSystem.Abs2MapCoord( Round( m_sx                         ), Round( m_sy                         ), xMapOR, yMapOR );
	MainSystem.Abs2MapCoord( Round( m_sx + x1 + CheckRect.Left   ), Round( m_sy + y1 + CheckRect.Bottom ), xMapDL, yMapDL );
	MainSystem.Abs2MapCoord( Round( m_sx + x1 + CheckRect.Right  ), Round( m_sy + y1 + CheckRect.Bottom ), xMapDR, yMapDR );
	MainSystem.Abs2MapCoord( Round( m_sx + x1 + CheckRect.Left   ), Round( m_sy + y1 + CheckRect.Top    ), xMapUL, yMapUL );
	MainSystem.Abs2MapCoord( Round( m_sx + x1 + CheckRect.Right  ), Round( m_sy + y1 + CheckRect.Top    ), xMapUR, yMapUR );

	if  MainSystem.TileMap.IsStandableMap( xMapOR, yMapOR, xMapDL, yMapDL ) and
		MainSystem.TileMap.IsStandableMap( xMapOR, yMapOR, xMapDR, yMapDR ) and
		MainSystem.TileMap.IsStandableMap( xMapOR, yMapOR, xMapUL, yMapUL ) and
		MainSystem.TileMap.IsStandableMap( xMapOR, yMapOR, xMapUR, yMapUR ) then begin

		m_sx := m_sx + x1;
		m_sy := m_sy + y1;

		m_x  := Round( m_sx );
		m_y  := Round( m_sy );

		MoveXY   := TRUE;
	end;
end;

procedure TPerson.MoveXYAuto;
begin

end;

procedure TPerson.MoveXYEx( Angle : Integer; Distance : Single; Intelligence : Boolean );

	function FindNearAngle( Angle : Integer; PreAngle : Integer; var dx, dy : Integer ) : Integer;
	const
		MAX_EXPECTED_ANGLE = 4;
		ExpectedAngle : array[0..Pred(MAX_EXPECTED_ANGLE)] of Integer = (
			0+25, 180-25, 180+25, 360-25
		);
	var
		i, Index, Diff : Integer;

	begin
		Index := -1;
		Diff  := 360;

		for i := 0 to Pred(MAX_EXPECTED_ANGLE) do begin
			if ( PreAngle >= 0 ) and ( PreAngle = ExpectedAngle[i] ) then
				continue;
			if ( Abs( Angle - ExpectedAngle[i] ) mod 360 ) < Diff then begin
				Index := i;
				Diff  := Abs( Angle - ExpectedAngle[i] ) mod 360;
			end;
		end;

		if Index >= 0 then begin
			dx := 0;
			dy := 0;

			case Index of
				0 : dx := 1;
				1 : dy := -1;
				2 : dx := -1;
				3 : dy := 1;
			end;

			Result := ExpectedAngle[Index];

		end else begin
			Result := 0;

		end;
	end;
var
	x1, y1           : Single;
	xOldMap, yOldMap : Integer;
	xNewMap, yNewMap : Integer;
	dx, dy, NewAngle : Integer;
begin
	x1 :=  SmCos( Angle ) * Distance;
	y1 := -SmSin( Angle ) * Distance;

	if not ( Intelligence ) then begin
		MoveXY( x1, y1 );
		Exit;
	end;

	MainSystem.Abs2MapCoord( Round( m_sx ), Round( m_sy ), xOldMap, yOldMap );
	MainSystem.Abs2MapCoord( Round( m_sx + x1 ), Round( m_sy + y1 ), xNewMap, yNewMap );

	if not MainSystem.TileMap.IsStandableMap( xOldMap, yOldMap, xNewMap, yNewMap ) then begin
		Exit;
		NewAngle := FindNearAngle( Angle, -1, dx, dy );
		if not MainSystem.TileMap.IsStandableMap( xOldMap, yOldMap, xOldMap + dx, yOldMap + dy ) then begin
			NewAngle := FindNearAngle( Angle, NewAngle, dx, dy );
			if not MainSystem.TileMap.IsStandableMap( xOldMap, yOldMap, xOldMap + dx, yOldMap + dy ) then begin
				Exit;
			end;
		end;
		x1 :=  SmCos( NewAngle ) * Distance;
		y1 := -SmSin( NewAngle ) * Distance;
	end;

	m_sx := m_sx + x1;
	m_sy := m_sy + y1;

	m_x  := Round( m_sx );
	m_y  := Round( m_sy );
end;

procedure TPerson.SavePersonToFile( var f : file );
begin

end;

function  TPerson.GetFaceImage : Integer;
begin
	if      m_direction < 90 then
		GetFaceImage := 3 * MAX_CHARA_FRAME
	else if m_direction < 180 then
		GetFaceImage := 2 * MAX_CHARA_FRAME
	else if m_direction < 270 then
		GetFaceImage := 0 * MAX_CHARA_FRAME
	else
		GetFaceImage := 1 * MAX_CHARA_FRAME
end;

procedure TPerson.SetFace( Direction : Integer );
begin
	m_direction := Direction;

	while m_direction < 0 do
		Inc( m_direction, 360 );

	m_direction := m_direction mod 360;
end;

function TPerson.GetFace : Integer;
begin
	GetFace := m_direction;
end;

{ TPersonManual }

procedure   TPersonManual.DisplaySelf( xDelta, yDelta : Integer; Selected : Boolean );
var
	DC : HDC;
	NewPen, OldPen : HPEN;
begin
	NewPen := CreatePen( PS_SOLID, 4, $F0FCFF );

	SmDirect3D.DirectDrawSurface.GetDC( DC );

	OldPen := SelectObject( DC, NewPen );

	MoveToEx( DC, m_x + xDelta + Round( 10 * SmCos( m_direction + 90 ) ), m_y + yDelta - Round( 10 * SmSin( m_direction + 90 ) ), nil );
	LineTo  ( DC, m_x + xDelta + Round( 50 * SmCos( m_direction ) ), m_y + yDelta - Round( 50 * SmSin( m_direction ) ) );
	LineTo  ( DC, m_x + xDelta + Round( 10 * SmCos( m_direction - 90 ) ), m_y + yDelta - Round( 10 * SmSin( m_direction - 90 ) ) );

	if ( OldPen > 0 ) then
		SelectObject( DC, OldPen );

	SmDirect3D.DirectDrawSurface.ReleaseDC( DC );

	DeleteObject( NewPen );

	Inherited;
end;

procedure   TPersonManual.DisplaySelfAfter( xDelta, yDelta : Integer; Selected : Boolean );
begin
	Inherited;
end;

function    TPersonManual.DoAction( IsMain :  Boolean ) : Boolean;
var
	InMoving : Boolean;
begin
	begin
		InMoving := FALSE;

		if HiByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_LEFT  ] ) ) > 0 then begin
			if HiByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_LOCK  ] ) ) > 0 then begin
				MoveXYEx( m_direction + 90, 1.5, TRUE );
			end else begin
				SetFace( m_direction + 3 );
			end;
			InMoving := TRUE;
		end;

		if HiByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_RIGHT ] ) ) > 0 then begin
			if HiByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_LOCK  ] ) ) > 0 then begin
				MoveXYEx( m_direction - 90, 1.5, TRUE );
			end else begin
				SetFace( m_direction - 3 );
			end;
			InMoving := TRUE;
		end;

		if HiByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_UP    ] ) ) > 0 then begin
			MoveXYEx( m_direction, 1.5, TRUE );
			InMoving := TRUE;
		end
		else if HiByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_DOWN  ] ) ) > 0 then begin
			MoveXYEx( m_direction + 180, 0.5, TRUE );
		end;

		if ( InMoving ) then begin
			if m_circle_scale < ( 256 + 100 ) then
				Inc( m_circle_scale, 2 );
		end else begin
			if m_circle_scale > 256 then
				Dec( m_circle_scale, 4 );
			if m_circle_scale < 256 then
				m_circle_scale := 256;
		end;

	end;

	DoAction := TRUE;
end;

{ TPersonManualLight }

function    TPersonManualLight.DoAction( IsMain :  Boolean ) : Boolean;
var
	x, y : Integer;
begin
	Inherited DoAction( IsMain );

	if LoByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_SHOOT  ] ) ) > 0 then begin
		MainSystem.CreateSprite( m_x-58-64, m_y+110+64+20, SmCos( m_direction ) * 8, - SmSin( m_direction ) * 8, 0, 0, ttTexture, 1, m_direction, 70 );
	end;
	if LoByte( GetAsyncKeyState( MainSystem.KeyList[0][SM_KEY_CHANGE ] ) ) > 0 then begin
		x := m_x-58-64;
		y := m_y+110+64+20;
		MainSystem.CreateSprite( x, y, SmCos( m_direction-20 ) * 4, - SmSin( m_direction-20 ) * 4, 0, 0, ttTexture, 1, m_direction-20, 70 );
		MainSystem.CreateSprite( x, y, SmCos( m_direction-10 ) * 4, - SmSin( m_direction-10 ) * 4, 0, 0, ttTexture, 1, m_direction-10, 70 );
		MainSystem.CreateSprite( x, y, SmCos( m_direction+00 ) * 4, - SmSin( m_direction+00 ) * 4, 0, 0, ttTexture, 1, m_direction+00, 70 );
		MainSystem.CreateSprite( x, y, SmCos( m_direction+10 ) * 4, - SmSin( m_direction+10 ) * 4, 0, 0, ttTexture, 1, m_direction+10, 70 );
		MainSystem.CreateSprite( x, y, SmCos( m_direction+20 ) * 4, - SmSin( m_direction+20 ) * 4, 0, 0, ttTexture, 1, m_direction+20, 70 );
	end;

	DoAction := TRUE;
end;

end.
